#include "Main.h"

Canvas::Canvas(Vector2n Position, bool Centered, bool HasBackground)
	: Widget(Position, Vector2n(0, 0)),
	  Camera(0, 0),
	  CameraZ(1),
	  m_Centered(Centered),
	  m_HasBackground(HasBackground),
	  m_Widgets()
{
}

Canvas::~Canvas()
{
}

void Canvas::Render()
{
	/*{
		MainCanvas.RenderBackground();

		MainCanvas.SetupTransform();
		{
			CurrentProject.Render();

			WidgetManager.Render(MainCanvas);
		}
		MainCanvas.EndTransform();
	}

	{
		MainCanvas2TEST.RenderBackground();

		MainCanvas2TEST.SetupTransform();
		{
			CurrentProject.Render();

			WidgetManager.Render(MainCanvas2TEST);
		}
		MainCanvas2TEST.EndTransform();
	}

	{
		OverlayCanvas.SetupTransform();
		{
			WidgetManager.Render(OverlayCanvas);

			g_TypingModuleTEST->Render();
		}
		OverlayCanvas.EndTransform();
	}*/

	if (m_HasBackground)
	{
		RenderBackground();
	}

	SetupTransform();
	{
		// HACK
		if (dynamic_cast<ConceptionApp *>(pMainApp) && m_HasBackground)
		{
			CurrentProject.Render();
		}

		for (auto & Widget : m_Widgets)
		{
			Widget.get()->Render();
		}
	}
	EndTransform();
}

void Canvas::AddWidget(Widget * Widget)
{
	m_Widgets.push_back(std::unique_ptr<class Widget>(Widget));
	Widget->SetParent(*this);
}

bool Canvas::HitTest(Vector2n Position, std::list<Widget *> * Hits) const
{
	bool InsideCanvas;
	
	if (m_HasBackground)
	{
		InsideCanvas = Widget::HitTest(Position, Hits);
	}
	else
	{
		InsideCanvas = Widget::HitTest(Position, nullptr);
	}

	bool HitInside = false;

	if (InsideCanvas)
	{
		Vector2d PositionDouble = GlobalToCanvas(Position);
		Vector2n PositionInt(static_cast<sint32>(std::lround(PositionDouble.X())), static_cast<sint32>(std::lround(PositionDouble.Y())));		// TODO: Loss of accuracy? Fix it if needed.

		//for (auto & Widget : m_Widgets)
		for (auto Widget = m_Widgets.rbegin(); m_Widgets.rend() != Widget; ++Widget)
		{
			if ((*Widget)->HitTest(PositionInt, Hits))
			{
				HitInside = true;
				break;		// HACK: Also a hack, need to do proper query if widget wants to block any futher hit testing for hover/interest checks
			}
		}
	}

	if (m_HasBackground)
	{
		return InsideCanvas;
	}
	else
	{
		return HitInside;
	}
}

#if 0
void Canvas::ProcessButton(Pointer * Pointer, Input::InputId ButtonId, bool Pressed)
{
	if (0 == ButtonId)
	{
		if (Pressed)
		{
			//Pointer->GetWidgetMapping().RequestActivation(this);
		}
		else
		{
			//Pointer->GetWidgetMapping().RequestDeactivation(this);
		}
	}
}

void Canvas::ProcessSlider(Pointer * Pointer, Input::InputId SliderId, double MovedAmount)
{
	// TODO: Make it based on this widget being active/having captured the pointer, rather than pointer being active (i.e. not exclusive)
	//if (Pointer->IsActive())
	//if (IsActiveExternally())
	//if (this == Pointer->GetWidgetMapping().GetActiveWidget())
	{
		if (   0 == SliderId
			|| 1 == SliderId)
		{
			MoveView(static_cast<uint8>(SliderId), MovedAmount);
		}
	}

	if (2 == SliderId)
	{
		/*double A[2] = { (GetAxisState(0).GetPosition() - 0.5 * GetAxisState(0).GetLength()),
						(GetAxisState(1).GetPosition() - 0.5 * GetAxisState(1).GetLength()) };*/
		double A[2] = { 0, 0 };
		MoveView(static_cast<uint8>(SliderId), MovedAmount, A);
	}
}
#endif

void Canvas::ProcessDrag(Vector2d DragAmount)
{
	double A[2] = { 0, 0 };
	MoveView(0, DragAmount[0], A);
	MoveView(1, DragAmount[1], A);
}

void Canvas::ProcessScroll(InputEvent & InputEvent, Vector2n ScrollAmount)
{
	Vector2n LocalPosition = Vector2n(InputEvent.GetPointerState().GetAxisState(0).GetPosition(), InputEvent.GetPointerState().GetAxisState(1).GetPosition()) - m_Position;
	double A[2] = { LocalPosition.X() - 0.5 * m_Dimensions.X(),
					LocalPosition.Y() - 0.5 * m_Dimensions.Y() };
	MoveView(2, ScrollAmount[0], A);
}

void Canvas::ProcessTimePassed(const double TimePassed)
{
	if (nullptr != m_GestureRecognizerTEST)
	{
		for (auto & Attached : m_GestureRecognizerTEST->GetAttached())
		{
			if (0 == Attached->GetDeviceCategory())
			{
				const PointerState & PointerState = Attached->GetPointerState();

				const double SpeedMultiplier = 250;

				if (PointerState.GetButtonState(GLFW_KEY_LEFT) && !PointerState.GetButtonState(GLFW_KEY_RIGHT))
				{
					static_cast<Canvas *>(pMainApp->m_Widgets.front().get())->MoveView(0, SpeedMultiplier * TimePassed);
				}
				else if (PointerState.GetButtonState(GLFW_KEY_RIGHT) && !PointerState.GetButtonState(GLFW_KEY_LEFT))
				{
					static_cast<Canvas *>(pMainApp->m_Widgets.front().get())->MoveView(0, -SpeedMultiplier * TimePassed);
				}

				if (PointerState.GetButtonState(GLFW_KEY_UP) && !PointerState.GetButtonState(GLFW_KEY_DOWN))
				{
					static_cast<Canvas *>(pMainApp->m_Widgets.front().get())->MoveView(1, SpeedMultiplier * TimePassed);
				}
				else if (PointerState.GetButtonState(GLFW_KEY_DOWN) && !PointerState.GetButtonState(GLFW_KEY_UP))
				{
					static_cast<Canvas *>(pMainApp->m_Widgets.front().get())->MoveView(1, -SpeedMultiplier * TimePassed);
				}
			}
		}
	}
}

void Canvas::ProcessCanvasUpdated()
{
	Widget::ProcessCanvasUpdated();

	for (auto & Widget : m_Widgets)
	{
		Widget->ProcessCanvasUpdated();
	}
}

void Canvas::MoveView(uint8 Degree, double MoveAmount, double A[2])
{
	switch (Degree)
	{
	case 0:
		Camera.X() -= MoveAmount / CameraZ;
		break;
	case 1:
		Camera.Y() -= MoveAmount / CameraZ;
		break;
	case 2:
		{
			if (MoveAmount > 0)
			{
				CameraZ *= 1.5;
				Camera.X() += A[0] * 0.5 / CameraZ;
				Camera.Y() += A[1] * 0.5 / CameraZ;
			}
			else if (MoveAmount < 0)
			{
				Camera.X() -= A[0] * 0.5 / CameraZ;
				Camera.Y() -= A[1] * 0.5 / CameraZ;
				CameraZ /= 1.5;
			}
		}
		break;
	default:
		break;
	}

	if (0 == Degree || 1 == Degree)
	{
		for (auto & Widget : m_Widgets)
		{
			Widget->ProcessCanvasUpdated();
		}
	}
}

void Canvas::RenderBackground()
{
	// HACK: Black background
	if (dynamic_cast<MultitouchTestApp *>(pMainApp) && m_HasBackground)
	{
		DrawBox(m_Position, m_Dimensions, Color(0, 0, 0), Color(1, 1, 1));

		return;
	}

	glBindTexture(GL_TEXTURE_2D, 2);

	/*glMatrixMode(GL_TEXTURE_MATRIX);
	glLoadIdentity();
	glScaled(CameraZ, CameraZ, CameraZ);
	glTranslated(-CameraX, -CameraY, 0);
	glMatrixMode(GL_MODELVIEW);*/

	glColor3d(1, 1, 1);
	glBegin(GL_QUADS);
		/*glTexCoord2d(0 / 256.0, 0 / 256.0); glVertex2d(-0.5 * m_CanvasSize[0], -0.5 * m_CanvasSize[1]);
		glTexCoord2d(0 / 256.0, m_CanvasSize[1] / 256.0); glVertex2d(-0.5 * m_CanvasSize[0], 0.5 * m_CanvasSize[1]);
		glTexCoord2d(m_CanvasSize[0] / 256.0, m_CanvasSize[1] / 256.0); glVertex2d(0.5 * m_CanvasSize[0], 0.5 * m_CanvasSize[1]);
		glTexCoord2d(m_CanvasSize[0] / 256.0, 0 / 256.0); glVertex2d(0.5 * m_CanvasSize[0], -0.5 * m_CanvasSize[1]);*/
		/*glTexCoord2d(-0.5 * m_CanvasSize[0] / 256.0 - 0.5, -0.5 * m_CanvasSize[1] / 256.0 - 0.5); glVertex2d(-0.5 * m_CanvasSize[0], -0.5 * m_CanvasSize[1]);
		glTexCoord2d(-0.5 * m_CanvasSize[0] / 256.0 - 0.5, 0.5 * m_CanvasSize[1] / 256.0 - 0.5); glVertex2d(-0.5 * m_CanvasSize[0], 0.5 * m_CanvasSize[1]);
		glTexCoord2d(0.5 * m_CanvasSize[0] / 256.0 - 0.5, 0.5 * m_CanvasSize[1] / 256.0 - 0.5); glVertex2d(0.5 * m_CanvasSize[0], 0.5 * m_CanvasSize[1]);
		glTexCoord2d(0.5 * m_CanvasSize[0] / 256.0 - 0.5, -0.5 * m_CanvasSize[1] / 256.0 - 0.5); glVertex2d(0.5 * m_CanvasSize[0], -0.5 * m_CanvasSize[1]);*/
		glTexCoord2d((-0.5 * m_Dimensions[0] + Camera.X() * CameraZ) / 256.0 / CameraZ - 0.5, (-0.5 * m_Dimensions[1] + Camera.Y() * CameraZ) / 256.0 / CameraZ - 0.5); glVertex2i(m_Position[0], m_Position[1]);
		glTexCoord2d((-0.5 * m_Dimensions[0] + Camera.X() * CameraZ) / 256.0 / CameraZ - 0.5, (0.5 * m_Dimensions[1] + Camera.Y() * CameraZ) / 256.0 / CameraZ - 0.5); glVertex2i(m_Position[0], m_Position[1] + m_Dimensions[1]);
		glTexCoord2d((0.5 * m_Dimensions[0] + Camera.X() * CameraZ) / 256.0 / CameraZ - 0.5, (0.5 * m_Dimensions[1] + Camera.Y() * CameraZ) / 256.0 / CameraZ - 0.5); glVertex2i(m_Position[0] + m_Dimensions[0], m_Position[1] + m_Dimensions[1]);
		glTexCoord2d((0.5 * m_Dimensions[0] + Camera.X() * CameraZ) / 256.0 / CameraZ - 0.5, (-0.5 * m_Dimensions[1] + Camera.Y() * CameraZ) / 256.0 / CameraZ - 0.5); glVertex2i(m_Position[0] + m_Dimensions[0], m_Position[1]);
	glEnd();
}

void Canvas::SetupTransform()
{
	glPushMatrix();

	glTranslated(m_Position.X(), m_Position.Y(), 0);
	if (m_Centered)
	{
		glTranslated(0.5 * m_Dimensions.X(), 0.5 * m_Dimensions.Y(), 0);
	}
	glScaled(CameraZ, CameraZ, CameraZ);
	glTranslated(-Camera.X(), -Camera.Y(), 0);

	glScissor(m_Position.X(), g_InputManager->GetWindowDimensions().Y() - (m_Position.Y() + m_Dimensions.Y()), m_Dimensions.X(), m_Dimensions.Y());
	glEnable(GL_SCISSOR_TEST);
}

void Canvas::EndTransform()
{
	glDisable(GL_SCISSOR_TEST);

	glPopMatrix();
}

const bool Canvas::GlobalIsInsideCanvas(const Vector2n Position) const
{
	return (   Position.X() >= m_Position.X()
			&& Position.Y() >= m_Position.Y()
			&& Position.X() <= m_Position.X() + m_Dimensions.X()
			&& Position.Y() <= m_Position.Y() + m_Dimensions.Y());
}

// Precondition: GlobalIsInsideCanvas(Position) is true
const Vector2d Canvas::GlobalToCanvas(const Vector2n Position) const
{
	Vector2d PositionDouble(Position.X(), Position.Y());

	if (!m_Centered)
	{
		return Vector2d((PositionDouble - Vector2d(m_Position.X(), m_Position.Y())) / CameraZ + Camera);
	}
	else
	{
		return Vector2d((PositionDouble - Vector2d(m_Position.X(), m_Position.Y()) - Vector2d(m_Dimensions.X(), m_Dimensions.Y()) * 0.5) / CameraZ + Camera);
	}
}

void Canvas::SetDimensions(Vector2n Dimensions)
{
	m_Dimensions = Dimensions;
}
