#include "../../Main.h"

Pointer::Pointer(InputResponder & MyInputResponder, const uint8 DeviceCategory)
	: InputResponder(),
	  m_GestureRecognizerTEST(nullptr),
	  m_DeviceCategory(DeviceCategory),
	  m_IsActive(false),
	  m_PointerState(),
	  //m_WidgetMapping()
	  m_InputResponder(MyInputResponder)
{
	//m_WidgetMapping.ProcessPointerCreation(this);
	InputEvent InputEvent;
	InputEvent.m_EventTypes.insert(InputEvent::EventType::POINTER_CREATION);
	InputEvent.m_DeviceCategory = m_DeviceCategory;
	InputEvent.m_Pointer = this;

	m_InputResponder.ProcessEvent(InputEvent);
}

Pointer::~Pointer()
{
	//m_WidgetMapping.ProcessPointerDeletion(this);
	InputEvent InputEvent;
	InputEvent.m_EventTypes.insert(InputEvent::EventType::POINTER_DELETION);
	InputEvent.m_DeviceCategory = m_DeviceCategory;
	InputEvent.m_Pointer = this;
	//InputEvent.m_Pointer = nullptr;

	m_InputResponder.ProcessEvent(InputEvent);
}

void Pointer::ProcessActivation(InputEvent & InputEvent)
{
	m_IsActive = true;

	//m_WidgetMapping.ProcessPointerActivation(this);
	InputEvent.m_EventTypes.insert(InputEvent::EventType::POINTER_ACTIVATION);
}

void Pointer::ProcessDeactivation(InputEvent & InputEvent)
{
	m_IsActive = false;

	//m_WidgetMapping.ProcessPointerDeactivation(this);
	InputEvent.m_EventTypes.insert(InputEvent::EventType::POINTER_DEACTIVATION);
}

bool Pointer::IsActive()
{
	return m_IsActive;
}

bool Pointer::ProcessEvent(InputEvent & InputEvent)
{
	//std::cout << InputEvent.ToString();

	for (auto EventType : InputEvent.m_EventTypes)
	{
		switch (EventType)
		{
		case InputEvent::EventType::BUTTON_EVENT:
			{
				auto ButtonId = InputEvent.m_InputId;
				bool Pressed = InputEvent.m_Buttons[0];

				// If no buttons were pressed thus far, then the pointer is getting activated
				if (Pressed && !InputEvent.GetPointerState().AnyButtonsPressed())
				{
					Pointer::ProcessActivation(InputEvent);
				}

				InputEvent.ModifyPointerState().UpdateButtonState(ButtonId) = Pressed;

				// If no buttons are pressed anymore, then the pointer is getting deactivated
				if (!Pressed && !InputEvent.GetPointerState().AnyButtonsPressed())
				{
					Pointer::ProcessDeactivation(InputEvent);
				}
			}
			break;
		case InputEvent::EventType::AXIS_EVENT:
			{
				for (Input::InputId AxisId = 0; AxisId < InputEvent.m_Axes.size(); ++AxisId)
				{
					int AxisPosition = InputEvent.m_Axes[AxisId].GetPosition();
					int PreviousAxisPosition = InputEvent.GetPointerState().GetAxisState(InputEvent.m_InputId + AxisId, InputEvent.m_Axes[AxisId]).GetPosition();
					int SliderValue = AxisPosition - PreviousAxisPosition;

					InputEvent.ModifyPointerState().UpdateAxisState(InputEvent.m_InputId + AxisId) = InputEvent.m_Axes[AxisId];

					InputEvent.m_Sliders.push_back(SliderValue);
				}
			}
			break;
		default:
			break;
		}
	}

	m_PointerState = InputEvent.GetPointerState();

	return m_InputResponder.ProcessEvent(InputEvent);
}

void Pointer::ProcessTimePassed(const double TimePassed)
{
}

/*WidgetMapping & Pointer::GetWidgetMapping()
{
	return m_WidgetMapping;
}*/

auto Pointer::GetPointerState() -> const PointerState &
{
	return m_PointerState;
}

auto Pointer::ModifyPointerState() -> PointerState &
{
	return m_PointerState;
}
